
// Panana Arduino part: reads usb serial and outputs to latch
// 30.08.2014: First vrs
// 31.08.2014: Trying to get Arduino play nice with all sorts of kludges
// 01.09.2014: Less kludges, more robust timing
// 01.09.2014: Save a second on final sync, replace writeDigital with
//             direct port commands

#define SYNC 60
#define HANDSHAKE 32

volatile unsigned char acked=1;

int timeout,syncci=SYNC,bytesread=0;

unsigned char prev=0,plus=0;

unsigned char flip[] ={
0,128,64,192,32,160,96,224,16,144,80,208,48,176,112,240,8,136,72,200,
40,168,104,232,24,152,88,216,56,184,120,248,4,132,68,196,36,164,100,
228,20,148,84,212,52,180,116,244,12,140,76,204,44,172,108,236,28,156,
92,220,60,188,124,252,2,130,66,194,34,162,98,226,18,146,82,210,50,178,
114,242,10,138,74,202,42,170,106,234,26,154,90,218,58,186,122,250,6,
134,70,198,38,166,102,230,22,150,86,214,54,182,118,246,14,142,78,206,
46,174,110,238,30,158,94,222,62,190,126,254,1,129,65,193,33,161,97,225,
17,145,81,209,49,177,113,241,9,137,73,201,41,169,105,233,25,153,89,217,
57,185,121,249,5,133,69,197,37,165,101,229,21,149,85,213,53,181,117,
245,13,141,77,205,45,173,109,237,29,157,93,221,61,189,125,253,3,131,
67,195,35,163,99,227,19,147,83,211,51,179,115,243,11,139,75,203,43,
171,107,235,27,155,91,219,59,187,123,251,7,135,71,199,39,167,103,231,
23,151,87,215,55,183,119,247,15,143,79,207,47,175,111,239,31,159,95,
223,63,191,127,255
};

void(* resetFunc) (void) = 0;

void setup()
{
  for(int i=4;i<=13;i++)
  {
    pinMode(i,OUTPUT);
    digitalWrite(i,LOW);
  }
  pinMode(3,INPUT);
  attachInterrupt(1,intte,RISING);
  
  digitalWrite(13,HIGH);
  Serial.begin(115200);
  while(Serial.read()!='@')
    ;
  while(Serial.read()!='#')
    ;
  Serial.write('!');
  digitalWrite(13,LOW);
  
  timeout=millis();
}

void loop()
{
  while(Serial.available()>0)
  {
    unsigned char c=Serial.read();
    
    if(bytesread<HANDSHAKE)
    {
      writebyte(c);
    }
    else // Following bytes need to be encoded
    {
      c+=plus; // Try to avoid repeating byte escapes
      plus++;
      
      if(c!=prev && c!=255) // The byte is the clock
      {
        writebyte(c);
      }
      else // Same byte or 0xff, needs encoding
      {
        writebyte(0xff);
        if(c>=0xfe)
        {
          writebyte(0xfe);
          if(c==0xfe)
          {
            writebyte(0);
            c=0;
          }
          else
          {
            writebyte(1);
            c=1;
          }
        }
        else
          writebyte(c);
      }
    }

    prev=c;

    syncci--;
    if(syncci==0)
    {
      syncci=SYNC;
      Serial.write('-');
    }
    
    timeout=millis();
    
    bytesread++;
  }
  
  if(millis()>timeout+100) // Transfer done
  {
    digitalWrite(13,HIGH);
    Serial.write('-');
    Serial.write('-');
    Serial.flush();
    Serial.end();
    detachInterrupt(1);
    resetFunc();
  }
}

void writebyte(unsigned char c)
{
  int tid;
  
  c=flip[c]; // Well, the pins happened to go this way :)
  
  // Yes, this is faster
  PORTD = (PORTD & 0x1f) | (c<<5);
  PORTB = (PORTB & 0xe0) | (c>>3);
  
  noInterrupts();
  PORTD &= 0xef; // Latch clock low
  acked=0;
  tid=millis();
  interrupts();

  while(acked==0)
  {
    // Timeout, let's get out of here!
    if(millis()>tid+20)
      acked=1;
  }
}

void intte()
{
  if(!acked)
  {
    acked=1;
    //digitalWrite(4,HIGH);
    PORTD |= 0x10;
  }
}
