
/* Panana PC transfer software by Marq/Fit

01.09.2014: The first version after deleting the friggin' source
06.09.2014: Switched to CJR from BIN
*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#define BPS B115200     // Serial speed
#define SYNC 60         // Sync after this many bytes
#define PROGRESS 1024   // Progress bar block size

// Default Arduino Uno port for Linux
#ifdef __linux__
#define PORT "/dev/ttyUSB0"
#endif

// Mac OS X
#ifdef __APPLE__
#define PORT "/dev/cu.usbmodem621"
#endif

// The Microsoft attempt at an OS (untested)
#ifdef __WIN32__
#define PORT "COM3"
#endif

// Sends a byte
void putbyte(int f,unsigned char c,int flush)
{
    static int syncci=SYNC;
    int i;
    unsigned char c2;

    i=write(f,&c,1);

    syncci--;
    if(!syncci || flush)
    {
        i=read(f,&c2,1);
        syncci=SYNC;
    }
}

int readcjr(char *name,int *addr,int *len,unsigned char *dada)
{
    FILE *f;
    int bytes=0,blknum=0,addy=0,
        n,i;

    f=fopen(name,"rb");
    if(f==NULL)
        return -1;

    while(1)
    {
        // Block header
        if(fgetc(f)==EOF)
            break;
        fgetc(f);
        blknum=fgetc(f);
        n=fgetc(f);
        if(n==0)
            n=256;

        if(blknum==1) // Code start address
        {
            addy=fgetc(f)*256;
            addy+=fgetc(f);
        }
        else
        {
            fgetc(f);
            fgetc(f);
        }

        // Hande the data
        for(i=0;i<n;i++)
        {
            if(blknum!=0 && blknum!=255)
            {
                dada[bytes]=fgetc(f);
                bytes++;
            }
            else
                fgetc(f);
        }

        // Checksum
        fgetc(f);
    }

    fclose(f);
    *len=bytes;
    *addr=addy;
    return 0;
}

int main(int argc,char *argv[])
{
    int f,i,c,
        size,addr;
    unsigned char checksum=0,c2;
    struct termios opt;
    char *portname=PORT;

    static unsigned char dada[32768];

    printf("Panana 0.01\n");

    if(argc<3)
    {
        printf("panana file.cjr autostart [port]\n");
        printf("autostart: 1/0\n");
        return(1);
    }

    // Open the binary
    if(readcjr(argv[1],&addr,&size,dada))
    {
        printf("File error.\n");
        return(1);
    }
    printf("File read. addr: $%.04x datalen: %d\n",addr,size);

    if(argc==4) // Override default port
        portname=argv[3];

    f=open(portname,O_RDWR|O_NOCTTY); // Open Arduino
    if(f==-1)
    {
        printf("Arduino can't be opened at %s\n",portname);
        return(1);
    }
    printf("Arduino opened\n");

    cfmakeraw(&opt); // Set serial params
    cfsetispeed(&opt,BPS);
    cfsetospeed(&opt,BPS);
    opt.c_cflag |= (CLOCAL | CREAD | CS8);
    opt.c_cflag &= ~(CRTSCTS);

    if(tcsetattr(f, TCSANOW, &opt)<0)
    {
        printf("Serial attribs can't be set\n");
        return(1);
    }

    // Handshake with Arduino
    printf("Waiting for Arduino\n");
    sleep(3);

    c2='@';
    for(i=0;i<50;i++)
    {
        c=write(f,&c2,1);
    }
    c2='#';
    c=write(f,&c2,1);
    do
    {
        i=read(f,&c2,1);
    }
    while(c2!='!');

    printf("Starting to transfer %d bytes",size);

    // Handshake with Pana
    for(i=0;i<30;i++)
        putbyte(f,0xff,0);
    putbyte(f,0x42,0);
    putbyte(f,0x43,0);

    // Transfer parameters
    putbyte(f,argv[2][0]-'0',0); // Autostart

    putbyte(f,size>>8,0);   // Data size
    putbyte(f,size&0xff,0);
    putbyte(f,addr>>8,0);   // Target address
    putbyte(f,addr&0xff,0);

    for(i=0;i<size;i++) // Byte by byte
    {
        c=dada[i];
        putbyte(f,c,0);
        checksum+=c;

        if((i&(PROGRESS-1))==0) // Progress bar
        {
            printf(".");
            fflush(stdout);
        }
    }

    putbyte(f,checksum,0);
    putbyte(f,0,1);

    printf("\nDone! Checksum %x\n",checksum);
    close(f);
    return(0);
}
